/*

    2002/05/23 - 1.4 - DeXT
    
    vcdisofs.c: subdirs now work for form2 files, too
    vcdisofs.c: form2 original extension can now be kept
    vcdisofs.c: fixed max filename length calculation (was greater than allowed)
    vcdisofs.c: ISO building is now faster for form1 files (no need to reserve space)
    
    2002/05/23 - 1.3 - DeXT
    
    vcdisofs.c: added subdirectories support (for Form1 files only)
    vcdisofs.c: added ASCII character set support

    2002/05/17 - 1.2 - DeXT
    
    vcdisofs.c: added Form1 files support

    2002/05/06 - 1.2pre2 - DeXT
    
    vcdisofs.c: improved XA compatibility (added XA sig for every file/dir)
    vcdisofs.c: added long file names support (ISO-9660 Level 2)
    vcdisofs.c: fixed emtpy block in ISO track
    
    2002/05/04 - 1.2pre - Nic
    
    vcdisofs.c: M2F2 file extension is no longer fixed

    2002/04/18 - beta 1.1 - DeXT
    
    vcdisofs.c: lowered ISO track size to 302 sectors (~700 Kb)

    2002/04/17 - beta 1.0 - DeXT

    vcdisofs.c: fixed make_path_tables() to not add root files as dirs
    vcdisofs.c: DAT files are added to the root directory instead of MPEGAV
    vcdisofs.c: removed unneeded VCD stuff

*/

/*
 * Make a (simple) ISO 9660 filesystem for a Video CD
 *
 * Many things in this program are just taken from mkisofs
 *
 * Structure of the filesystem
 *
 *     Block      Content
 *
 *     0-15       empty
 *     16         Primary Volume descriptor
 *     17         End Volume descriptor
 *     18         Path table (Intel Byte order) (must fit in 1 Sector)
 *     19         Path table (Motorola Byte order) (must fit in 1 Sector)
 *     20         Root directory
 *     21  ...    Other directories
 *     150 ...    Files in directory VCD
 *     210 ...    Other files
 */

#ifdef _WIN32
#include <windows.h>
#endif
#include <stdio.h>
#include <time.h>
#include "defaults.h"

/* Path table stuff. Note that this is simplified and only valid
   if there are no subdirectories below the directories in the
   root directory and if the path tables fit into 1 sector */

#define PVD_EXTENT 16
#define EVD_EXTENT 17
#define PATH_TABLE_L_EXTENT 18
#define PATH_TABLE_M_EXTENT 19
#define ROOT_DIR_EXTENT   20

#define ISO_DIR_SIZE 2048

static char path_table_l[2048];
static char path_table_m[2048];
static int  path_table_size;

static int root_dir_len;
static char root_dir[ISO_DIR_SIZE];
static int cur_dir_extent;
static int cur_file_extent;

static char zero2048[2048] = { 0, };
static char buff2048[2048];

/* Various files on the VCD */

static char entries_file[2048];
static char info_file[2048] = {
  'V', 'I', 'D', 'E', 'O', '_', 'C', 'D',
    1,   1, '1', ' ', ' ', ' ', ' ', ' ',
  ' ', ' ', ' ', ' ', ' ', ' ', ' ', ' ',
  ' ', ' ',   0,   1,   0,   1,   0,   0,
};

static struct tm *t;

#define BCD(x) ( ((x)/10)*16 + (x)%10 )

static void set_721( char *pnt, unsigned int i)
{
     pnt[0] = i & 0xff;
     pnt[1] = (i >> 8) &  0xff;
}

static void set_722( char *pnt, unsigned int i)
{
     pnt[0] = (i >> 8) &  0xff;
     pnt[1] = i & 0xff;
}

static void set_723( char *pnt, unsigned int i)
{
     pnt[3] = pnt[0] = i & 0xff;
     pnt[2] = pnt[1] = (i >> 8) &  0xff;
}

static void set_731( char *pnt, unsigned int i)
{
     pnt[0] = i & 0xff;
     pnt[1] = (i >> 8) &  0xff;
     pnt[2] = (i >> 16) &  0xff;
     pnt[3] = (i >> 24) &  0xff;
}

static void set_732( char *pnt, unsigned int i)
{
     pnt[3] = i & 0xff;
     pnt[2] = (i >> 8) &  0xff;
     pnt[1] = (i >> 16) &  0xff;
     pnt[0] = (i >> 24) &  0xff;
}

static void set_733( char *pnt, unsigned int i)
{
     pnt[7] = pnt[0] = i & 0xff;
     pnt[6] = pnt[1] = (i >> 8) &  0xff;
     pnt[5] = pnt[2] = (i >> 16) &  0xff;
     pnt[4] = pnt[3] = (i >> 24) &  0xff;
}

static void set_str( char *pnt, char *src, unsigned int len)
{
     int i, l;
     l = strlen(src);

     for(i=0;i<len;i++) pnt[i] = (i<l) ? src[i] : ' ';
}


void make_path_tables()
{
   int i, j, len, extent;
   unsigned char *root_dir_entry;

   memset(path_table_l, 0, sizeof(path_table_l));
   memset(path_table_m, 0, sizeof(path_table_m));

   path_table_size = 0;

   for(root_dir_entry = root_dir;
       root_dir_entry[0]>0;
       root_dir_entry += root_dir_entry[0] )
   {
      /* Skip the .. entry in the root directory */

      if (root_dir_entry[33]==1) continue;
      
      if (root_dir_entry[25]==0) continue;  // Do not add File Records to Path Table!

      len = root_dir_entry[32];
      path_table_l[path_table_size] = len;
      path_table_m[path_table_size] = len;
      path_table_size += 2;

      extent =  root_dir_entry[2]      +
               (root_dir_entry[3]<< 8) +
               (root_dir_entry[4]<<16) +
               (root_dir_entry[5]<<24);
      set_731(path_table_l + path_table_size, extent);
      set_732(path_table_m + path_table_size, extent);
      path_table_size += 4;

      set_721(path_table_l + path_table_size, 1);
      set_722(path_table_m + path_table_size, 1);
      path_table_size += 2;

      for(j=0; j<len; j++)
      {
         path_table_l[path_table_size] = root_dir_entry[33+j];
         path_table_m[path_table_size] = root_dir_entry[33+j];
         path_table_size++;
      }
      if(path_table_size & 1) path_table_size++;
   }
}

/* From mkisofs: */

#define ISODCL(from, to) (to - from + 1)

static struct iso_primary_descriptor {
        char type                       [ISODCL (  1,   1)]; /* 711 */
        char id                         [ISODCL (  2,   6)];
        char version                    [ISODCL (  7,   7)]; /* 711 */
        char unused1                    [ISODCL (  8,   8)];
        char system_id                  [ISODCL (  9,  40)]; /* achars */
        char volume_id                  [ISODCL ( 41,  72)]; /* dchars */
        char unused2                    [ISODCL ( 73,  80)];
        char volume_space_size          [ISODCL ( 81,  88)]; /* 733 */
        char escape_sequences           [ISODCL ( 89, 120)];
        char volume_set_size            [ISODCL (121, 124)]; /* 723 */
        char volume_sequence_number     [ISODCL (125, 128)]; /* 723 */
        char logical_block_size         [ISODCL (129, 132)]; /* 723 */
        char path_table_size            [ISODCL (133, 140)]; /* 733 */
        char type_l_path_table          [ISODCL (141, 144)]; /* 731 */
        char opt_type_l_path_table      [ISODCL (145, 148)]; /* 731 */
        char type_m_path_table          [ISODCL (149, 152)]; /* 732 */
        char opt_type_m_path_table      [ISODCL (153, 156)]; /* 732 */
        char root_directory_record      [ISODCL (157, 190)]; /* 9.1 */
        char volume_set_id              [ISODCL (191, 318)]; /* dchars */
        char publisher_id               [ISODCL (319, 446)]; /* achars */
        char preparer_id                [ISODCL (447, 574)]; /* achars */
        char application_id             [ISODCL (575, 702)]; /* achars */
        char copyright_file_id          [ISODCL (703, 739)]; /* 7.5 dchars */
        char abstract_file_id           [ISODCL (740, 776)]; /* 7.5 dchars */
        char bibliographic_file_id      [ISODCL (777, 813)]; /* 7.5 dchars */
        char creation_date              [ISODCL (814, 830)]; /* 8.4.26.1 */
        char modification_date          [ISODCL (831, 847)]; /* 8.4.26.1 */
        char expiration_date            [ISODCL (848, 864)]; /* 8.4.26.1 */
        char effective_date             [ISODCL (865, 881)]; /* 8.4.26.1 */
        char file_structure_version     [ISODCL (882, 882)]; /* 711 */
        char unused4                    [ISODCL (883, 883)];
        char application_data           [ISODCL (884, 1395)];
        char unused5                    [ISODCL (1396, 2048)];
} ipd;

static struct iso_directory_record {
        unsigned char length            [ISODCL (1, 1)];   /* 711 */
        char ext_attr_length            [ISODCL (2, 2)];   /* 711 */
        char extent                     [ISODCL (3, 10)];  /* 733 */
        char size                       [ISODCL (11, 18)]; /* 733 */
        char date                       [ISODCL (19, 25)]; /* 7 by 711 */
        char flags                      [ISODCL (26, 26)];
        char file_unit_size             [ISODCL (27, 27)]; /* 711 */
        char interleave                 [ISODCL (28, 28)]; /* 711 */
        char volume_sequence_number     [ISODCL (29, 32)]; /* 723 */
        unsigned char name_len          [ISODCL (33, 33)]; /* 711 */
        char name                       [34]; /* Not really, but we need something here */
} idr;


static void add_dirent(char *dir, int *dirlen, int dirsize,
           char *name, int namelen, int extent, int size, int flags, int xa)
{
   int reclen;
   unsigned char xa_attr[2];

   reclen = 33 + namelen;
   if(reclen & 1) reclen++;

   if(xa >= 0) reclen += 14; // add XA signature

   if( (*dirlen) + reclen > dirsize )
   {
      fprintf(stderr,"Fatal Error: add_dirent - max dirsize exceeded\n");
      exit(1);
   }

   memset(dir+(*dirlen), 0, reclen);

   memset(&idr, 0 , sizeof(idr));

   idr.length[0] = reclen;
   idr.ext_attr_length[0] = 0;
   set_733(idr.extent,extent);
   set_733(idr.size,size);

   idr.date[0] = t->tm_year;
   idr.date[1] = t->tm_mon+1;
   idr.date[2] = t->tm_mday;
   idr.date[3] = t->tm_hour;
   idr.date[4] = t->tm_min;
   idr.date[5] = t->tm_sec;
   idr.date[6] = 0; /* This is in units of 15 minutes from GMT */

   idr.flags[0] = flags;
   idr.file_unit_size[0] = 0;
   idr.interleave[0] = 0;

   set_723(idr.volume_sequence_number,1);

   idr.name_len[0] = namelen;

   memcpy(dir+(*dirlen), &idr, 33);
   memcpy(dir+(*dirlen)+33, name, namelen);

   /* XA signature */

   if (xa >= 0) {
      if (flags == 2) {
         xa_attr[0] = 0x8d;  // directory
         xa_attr[1] = 0x55;
      } else {
         if (xa == 0) {
            xa_attr[0] = 0x0d;  // form1 file
            xa_attr[1] = 0x55;
         } else {
            xa_attr[0] = 0x15;  // form2 file
            xa_attr[1] = 0x55;
         }
      }

      dir[*dirlen+reclen-10] = xa_attr[0];
      dir[*dirlen+reclen- 9] = xa_attr[1];
      dir[*dirlen+reclen- 8] = 'X';
      dir[*dirlen+reclen- 7] = 'A';
      dir[*dirlen+reclen- 6] = xa;
   }

   (*dirlen) += reclen;
}

static void make_ipd()
{
   char iso_time[17];
   int len;
   char name[1];

   memset(&ipd,0,sizeof(ipd));

   ipd.type[0] = 1;
   set_str(ipd.id,"CD001",5);
   ipd.version[0] = 1;

   set_str(ipd.system_id,CD_SYSTEM_ID,32);
   set_str(ipd.volume_id,CD_VOLUME_ID,32);

   set_733(ipd.volume_space_size,ISO_FS_BLOCKS);
   /* we leave ipd.escape_sequences = 0 */

   set_723(ipd.volume_set_size,1);
   set_723(ipd.volume_sequence_number,1);
   set_723(ipd.logical_block_size,2048);

   set_733(ipd.path_table_size,path_table_size);
   set_731(ipd.type_l_path_table,PATH_TABLE_L_EXTENT);
   set_731(ipd.opt_type_l_path_table,0);
   set_732(ipd.type_m_path_table,PATH_TABLE_M_EXTENT);
   set_732(ipd.opt_type_m_path_table,0);

   len = 0;
   name[0] = 0;
   add_dirent(ipd.root_directory_record, &len, 34,
              name, 1, ROOT_DIR_EXTENT, ISO_DIR_SIZE, 2, -1);
   if(len!=34)
   {
      fprintf(stderr,"Internal error in make_ipd\n");
      exit(1);
   }

   set_str(ipd.volume_set_id , CD_VOLUME_SET_ID , 128);
   set_str(ipd.publisher_id  , CD_PUBLISHER_ID  , 128);
   set_str(ipd.preparer_id   , CD_PREPARER_ID   , 128);
   set_str(ipd.application_id, CD_APPLICATION_ID, 128);

   set_str(ipd.copyright_file_id    , " ", 37);
   set_str(ipd.abstract_file_id     , " ", 37);
   set_str(ipd.bibliographic_file_id, " ", 37);

   sprintf(iso_time, "%4.4d%2.2d%2.2d%2.2d%2.2d%2.2d00",
              1900+(t->tm_year), t->tm_mon+1, t->tm_mday,
              t->tm_hour, t->tm_min, t->tm_sec);
   iso_time[16] = 0; /* This is in binary! Units of 15 minutes from GMT */

   memcpy(ipd.creation_date,    iso_time,17);
   memcpy(ipd.modification_date,iso_time,17);
   memcpy(ipd.expiration_date,  "0000000000000000\0",17);
   memcpy(ipd.effective_date,   iso_time,17);

   ipd.file_structure_version[0] = 1;

   /*  The string CD-XA001 in this exact position indicates  */
   /*  that the disk contains XA sectors.                    */
   memcpy(ipd.application_data+141, "CD-XA001", 8);
}

static void init_iso_dir(char *dir, int *dirlen, int self, int parent)
{
   char name[1];

   memset(dir,0,ISO_DIR_SIZE);
   *dirlen = 0;

   /* add . and .. entries */

   name[0] = 0;
   add_dirent(dir, dirlen, ISO_DIR_SIZE, name, 1, self,   ISO_DIR_SIZE, 2, 0);

   name[0] = 1;
   add_dirent(dir, dirlen, ISO_DIR_SIZE, name, 1, parent, ISO_DIR_SIZE, 2, 0);
}

void add_CDI_dir()
{
   char dir[ISO_DIR_SIZE];
   int dirlen, len, i;

   cur_dir_extent++;

   if(cur_dir_extent>=START_FILE_EXTENT)
   { fprintf(stderr,"Fatal Error: too many dirs\n"); exit(1); }

   /* Initialize directory */

   init_iso_dir(dir, &dirlen, cur_dir_extent, ROOT_DIR_EXTENT);

   /* Add this directory to root directory */

   add_dirent(root_dir, &root_dir_len, ISO_DIR_SIZE,
              "CDI", 3, cur_dir_extent, ISO_DIR_SIZE, 2, 0);

   /* Add files */

   /* Since this directory is only used for CDI players
      which are hardly in use any more, we add only a bogus file
      CDI_VCD.APP consisting of 0's */

   /* CDI_VCD.APP */

   len = 2048;
   add_dirent(dir, &dirlen, ISO_DIR_SIZE,
              "CDI_VCD.APP;1", 13, cur_file_extent, len, 0, 0);
   for(i=0;i<LEN2BLOCKS(len);i++) output_form1(cur_file_extent++,zero2048);

   /* Output the directory record */

   output_form1(cur_dir_extent,dir);
}

void add_new_dir(char *dir, int *dirlen, char *name)
{
   cur_dir_extent++;

   if(cur_dir_extent>=START_FILE_EXTENT)
   { fprintf(stderr,"Fatal Error: too many dirs\n"); exit(1); }

   /* Initialize directory */

   init_iso_dir(dir, dirlen, cur_dir_extent, ROOT_DIR_EXTENT);

   /* Add this directory to root directory */

   add_dirent(root_dir, &root_dir_len, ISO_DIR_SIZE,
              name, strlen(name), cur_dir_extent, ISO_DIR_SIZE, 2, 0);
}

void to_dchar(char *dest, char *src, int max)
{
int j;

   for (j = 0; j < (strlen(src) <= max ? strlen(src) : max); j++)
       dest[j] = (isalnum(src[j]) ? toupper(src[j]) : '_'); // d-char (ISO-9660)
   dest[j] = 0;
}

void convert_file_name(char *dstname, char *srcname, int form, int long_file_names, int ascii, int form2_ext_to_name)
{
char *filename, *ext;
char FORM1_EXT[260] = { 0, };

      if (!ascii) to_dchar(FORM2_EXT, FORM2_EXT, 30);
      
         filename = (char *) strrchr(srcname, DIR_SLASH);

         if (filename == NULL)
            filename = srcname;
         else
            filename = &filename[1];  // skip directory slash

         strcpy(dstname, filename);

         ext = (char *) strrchr(dstname, '.');
         
         if (ext != NULL)
         {
            if (form == FORM1)
            {
               strcpy(FORM1_EXT,&ext[1]);
               if (!ascii) to_dchar(FORM1_EXT, FORM1_EXT, 30);

               ext[0] = 0; // trim the file extension, if any
            }
            else // FORM2
               if (!form2_ext_to_name) ext[0] = 0; // trim the file extension, if any
         }   

         if (!long_file_names)
         {
            if (strlen(dstname) > 8) dstname[8] = 0;
            if (strlen(FORM1_EXT) > 3) FORM1_EXT[3] = 0;
            if (strlen(FORM2_EXT) > 3) FORM2_EXT[3] = 0;
         }

         if (!ascii)
            to_dchar(dstname, dstname, 30-(strlen(form == FORM1 ? FORM1_EXT : FORM2_EXT)));
         
         strcat(dstname,"."); // separator is mandatory
         
         if (form == FORM2)
            strcat(dstname,FORM2_EXT);
         else
            if (ext != NULL) strcat(dstname,FORM1_EXT);
         
         strcat(dstname,";1"); // file version number
}

void convert_dir_name(char *dstname, char *srcname, int long_file_names, int ascii)
{
         strcpy(dstname, srcname);

         if (!long_file_names)
         {
            if (strlen(dstname) > 8) dstname[8] = 0;
         }

         if (!ascii) to_dchar(dstname, dstname, 31);  // extensions not allowed in ISO-9660
}

void mk_vcd_iso_fs(file_s *file, int num_entries, int num_dirs, int long_file_names, int ascii, int form2_ext_to_name)
{
   int i, n, len;
   time_t T;
   char name[260];

   char *cur_dir_ptr;
   int *cur_dir_len_ptr;
   char cur_dir[MAX_DIRS][ISO_DIR_SIZE]; // TODO: allocate memory at runtime
   int cur_dir_len, cur_dir_order;
   char dirname[260] = { 0, };

   /* some initializations */

   time(&T);
   t = gmtime(&T);

   init_iso_dir(root_dir, &root_dir_len, ROOT_DIR_EXTENT, ROOT_DIR_EXTENT);

   cur_dir_extent = ROOT_DIR_EXTENT;
   cur_file_extent = START_FILE_EXTENT;

   /* set the first 16 blocks to 0 */

   for(i=0;i<PVD_EXTENT;i++) output_form1(i,zero2048);

   /* Output the data */

   cur_dir_order = 0;
   cur_dir_ptr = root_dir;
   cur_dir_len_ptr = &root_dir_len;

   for(n=0;n<num_entries;n++)
   {
      if (file[n].type == DIR) // new dir
      {
         cur_dir_len_ptr = &cur_dir_len;
         cur_dir_ptr = cur_dir[cur_dir_order++];

         convert_dir_name(dirname, file[n].name, long_file_names, ascii);

         printf("%s\\\n",dirname);

         add_new_dir(cur_dir_ptr, cur_dir_len_ptr, dirname);
      }
      else if (file[n].type != NOTFILE) // files
      {
         convert_file_name(name, file[n].name, file[n].type, long_file_names, ascii, form2_ext_to_name);

         printf("%s\\%s\n",dirname,name);

         if (file[n].type == FORM1)
            len = file[n].length;
         else // FORM2
            len = file[n].size*2048; // dummy size for FORM2 files

         add_dirent(cur_dir_ptr, cur_dir_len_ptr, ISO_DIR_SIZE,
                   name, strlen(name), file[n].extent, len, 0, file[n].number);
      }
   }

   if (num_dirs > 0) // output dir entries
   {
      for(i = 0; i < num_dirs; i++)
         output_form1(i+ROOT_DIR_EXTENT+1, cur_dir[i]);
   }

   /* Fill the rest of the ISO FS */

   for(i=cur_dir_extent+1;i<START_FILE_EXTENT;i++)
      output_form1(i,zero2048);

   /* Create and output path tables */

   make_path_tables();

   output_form1(PATH_TABLE_L_EXTENT,path_table_l);
   output_form1(PATH_TABLE_M_EXTENT,path_table_m);

   /* Output root directory */

   output_form1(ROOT_DIR_EXTENT, root_dir);

   /* Create ISO Primary desriptor */

   make_ipd();
   output_form1(PVD_EXTENT,(char *) &ipd);

   /* Create end volume descriptor */

   memset(buff2048, 0, 2048);

   buff2048[0] = 0xff;
   buff2048[1] = 'C';
   buff2048[2] = 'D';
   buff2048[3] = '0';
   buff2048[4] = '0';
   buff2048[5] = '1';
   buff2048[6] = 0x01;
   output_form1(EVD_EXTENT, buff2048);
}

